"""This subpackage contains generic utilities for fitting, parsing
CHESS scan data, collecting interactive user input, and finding
lattice properties of materials (among others).
"""

from CHAP.utils.fit import FitProcessor
